@@ -10,7 +10,6 @@ import gzip
 import itertools
 import json
 import logging
-import sys
 import time
 import traceback
 import uuid
@@ -158,6 +157,7 @@ EXTERNAL_BROWSER_AUTHENTICATOR = 'EXTERNALBROWSER'
 KEY_PAIR_AUTHENTICATOR = 'SNOWFLAKE_JWT'
 OAUTH_AUTHENTICATOR = 'OAUTH'
 ID_TOKEN_AUTHENTICATOR = 'ID_TOKEN'
+USR_PWD_MFA_AUTHENTICATOR = 'USERNAME_PASSWORD_MFA'
 
 
 def is_retryable_http_code(code: int) -> bool:
@@ -248,12 +248,20 @@ class SnowflakeRestful(object):
 
     @property
     def id_token(self):
-        return self._id_token if hasattr(self, '_id_token') else None
+        return getattr(self, '_id_token', None)
 
     @id_token.setter
     def id_token(self, value):
         self._id_token = value
 
+    @property
+    def mfa_token(self):
+        return getattr(self, '_mfa_token', None)
+
+    @mfa_token.setter
+    def mfa_token(self, value):
+        self._mfa_token = value
+
     def close(self):
         if hasattr(self, '_token'):
             del self._token
@@ -261,6 +269,8 @@ class SnowflakeRestful(object):
             del self._master_token
         if hasattr(self, '_id_token'):
             del self._id_token
+        if hasattr(self, '_mfa_token'):
+            del self._mfa_token
         sessions = list(self._active_sessions)
         if sessions:
             logger.debug("Closing %s active sessions", len(sessions))
@@ -313,12 +323,14 @@ class SnowflakeRestful(object):
 
     def update_tokens(self, session_token, master_token,
                       master_validity_in_seconds=None,
-                      id_token=None):
+                      id_token=None,
+                      mfa_token=None):
         """Updates session and master tokens and optionally temporary credential."""
         with self._lock_token:
             self._token = session_token
             self._master_token = master_token
             self._id_token = id_token
+            self._mfa_token = mfa_token
             self._master_validity_in_seconds = master_validity_in_seconds
 
     def _renew_session(self):
@@ -327,9 +339,7 @@ class SnowflakeRestful(object):
 
     def _token_request(self, request_type):
         logger.debug(
-            'updating session. master_token: %s, id_token: %s',
-            '****' if self.master_token else None,
-            '****' if self.id_token else None)
+            'updating session. master_token: {}'.format('****' if self.master_token else None))
         headers = {
             HTTP_HEADER_CONTENT_TYPE: CONTENT_TYPE_APPLICATION_JSON,
             HTTP_HEADER_ACCEPT: CONTENT_TYPE_APPLICATION_JSON,
@@ -359,8 +369,7 @@ class SnowflakeRestful(object):
                 ret['data']['sessionToken'],
                 ret['data'].get('masterToken'),
                 master_validity_in_seconds=ret['data'].get(
-                    'masterValidityInSeconds'),
-                id_token=self.id_token)
+                    'masterValidityInSeconds'))
             logger.debug('updating session completed')
             return ret
         else:
@@ -638,7 +647,6 @@ class SnowflakeRestful(object):
             return {}
         except RetryRequest as e:
             if retry_ctx.cnt == TelemetryService.get_instance().num_of_retry_to_trigger_telemetry:
-                _, _, stack_trace = sys.exc_info()
                 TelemetryService.get_instance().log_http_request_error(
                     "HttpRequestRetry%dTimes" % retry_ctx.cnt,
                     full_url,
@@ -657,7 +665,6 @@ class SnowflakeRestful(object):
                 retry_ctx.timeout -= int(time.time() - start_request_thread)
                 if retry_ctx.timeout <= 0:
                     logger.error(cause, exc_info=True)
-                    _, _, stack_trace = sys.exc_info()
                     TelemetryService.get_instance().log_http_request_error(
                         "HttpRequestRetryTimeout",
                         full_url,
@@ -872,7 +879,6 @@ class SnowflakeRestful(object):
                 exc_info=True)
             raise RetryRequest(err)
         except Exception as err:
-            _, _, stack_trace = sys.exc_info()
             TelemetryService.get_instance().log_http_request_error(
                 "HttpException%s" % str(err),
                 full_url,
